/*
 Storybook: Open Source software for novelists and authors.
 Copyright (C) 2008 - 2012 Martin Mustun

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package storybook.toolkit;

import i18n.I18N;
import java.awt.event.ActionEvent;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import net.infonode.docking.DockingWindow;
import net.infonode.docking.RootWindow;
import net.infonode.docking.TabWindow;
import net.infonode.docking.View;
import net.infonode.docking.WindowBar;
import net.infonode.docking.properties.TabWindowProperties;
import net.infonode.docking.util.StringViewMap;
import resources.MainResources;
import storybook.App;
import storybook.Pref;
import storybook.ctrl.Ctrl;
import storybook.toolkit.file.EnvUtil;
import storybook.ui.MainFrame;
import storybook.ui.SbView;
import storybook.ui.SbViewFactory;
import storybook.ui.dialog.ExceptionDlg;
import storybook.ui.panel.AbstractPanel;

/**
 * @author martin
 *
 */
public class DockUtil {

	private static final String TT = "DockUtil",
			LAYOUT_EXT = ".layout";

	private DockUtil() {
		// nothing
	}

	public static void setRespectMinimumSize(MainFrame mainFrame) {
		SbView view = mainFrame.getView(SbView.VIEWNAME.TREE);
		DockingWindow win = view.getWindowParent();
		if (win instanceof TabWindow) {
			TabWindowProperties props = new TabWindowProperties();
			props.setRespectChildWindowMinimumSize(true);
			((TabWindow) win).getTabWindowProperties().addSuperObject(props);
		} else if (win instanceof WindowBar) {
			// can't be set for a WindowBar
		}
	}

	public static void layoutSet(MainFrame mainFrame, String hex) {
		mainFrame.cursorSetWaiting();
		byte[] ba = hex.getBytes();
		try ( ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(ba))) {
			RootWindow rootWindow = mainFrame.getRootWindow();
			rootWindow.read(in, true);
		} catch (NullPointerException | IOException e) {
			ExceptionDlg.show(I18N.getMsg("err.layout_reading"), e);
			return;
		}
		setRespectMinimumSize(mainFrame);
		unloadHiddenViews(mainFrame);
		mainFrame.cursorSetDefault();
	}

	public static void unloadHiddenViews(MainFrame mainFrame) {
		SbViewFactory viewFactory = mainFrame.getViewFactory();
		StringViewMap viewMap = viewFactory.getViewMap();
		Ctrl documentCtrl = mainFrame.getBookController();
		int c = viewMap.getViewCount();
		for (int i = 0; i < c; ++i) {
			View view = viewMap.getViewAtIndex(i);
			if (view instanceof SbView) {
				SbView sbView = (SbView) view;
				if (sbView.isWindowShowing()) { // window is showing
					if (!sbView.isLoaded()) {
						viewFactory.loadView(sbView);
						documentCtrl.attachView((AbstractPanel) view.getComponent());
					}
				} else // window is not showing
				if (sbView.isLoaded()) {
					documentCtrl.detachView((AbstractPanel) view.getComponent());
					viewFactory.unloadView(sbView);
				}
			}
		}
	}

	public static void layoutSave(MainFrame mainFrame, String name) {
		try {
			RootWindow rootWindow = mainFrame.getRootWindow();
			if (rootWindow == null) {
				return;
			}
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			try ( ObjectOutputStream out = new ObjectOutputStream(bos)) {
				rootWindow.write(out, false);
			}
			byte[] ba = bos.toByteArray();
			File f = new File(EnvUtil.getPrefDir().getAbsolutePath()
					+ File.separator + name + LAYOUT_EXT);
			try ( FileOutputStream fos = new FileOutputStream(f)) {
				fos.write(ba);
			}
			List<String> list = App.getInstance().preferences.dockingGetList();
			list.add(name);
			App.getInstance().preferences.dockingSetList(Pref.KEY.DOCKING_LAYOUT, list);
			App.getInstance().preferences.setString(Pref.KEY.LAYOUT_LAST_USED, name);
			App.getInstance().reloadMenuBars();
			App.getInstance().reloadStatusBars();
		} catch (IOException e) {
			LOG.err("*** " + TT + ".saveLoayout(" + mainFrame.getName()
					+ "," + name + ") Exception", e);
		}
	}

	public static boolean layoutLoad(MainFrame mainFrame, String name) {
		//App.trace("DockUtil.loadLayout(mainFrame, name="+name+")");
		try {
			Path path = Paths.get(EnvUtil.getPrefDir().getAbsolutePath()
					+ File.separator + name + LAYOUT_EXT);
			if (!path.toFile().exists()) {
				LOG.err("Layout not find: " + path.toString());
				return (false);
			}
			byte[] ba = Files.readAllBytes(path);
			try ( ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(ba))) {
				RootWindow rootWindow = mainFrame.getRootWindow();
				rootWindow.read(in, true);
			} catch (Exception ex) {
				LOG.err("unable to restore layout, setting default");
				layoutLoadFile(mainFrame, "default");
				return (false);
			}
			mainFrame.cursorSetWaiting();
			setRespectMinimumSize(mainFrame);
			unloadHiddenViews(mainFrame);
			mainFrame.cursorSetDefault();
			return true;
		} catch (IOException ex) {
			LOG.err("*** " + TT + ".loadLoayout(" + mainFrame.getName()
					+ "," + name + ") Exception :" + ex.getMessage());
		}
		return false;
	}

	public static JMenu layoutGetMenu(MainFrame mainFrame, JMenu menu) {
		String[] list = {
			"default",
			"only_book",
			"only_chrono",
			"only_manage",
			"only_reading",
			"persons_locations",
			"tags_items",
			"screenplay",
			"theaterplay",
			"storyboard"
		};
		menu.removeAll();
		for (String key : list) {
			JMenuItem m = new JMenuItem();
			m.setFont(App.getInstance().fontGetDefault());
			m.setName(key);
			m.setText(I18N.getMsg("docking.layout." + key));
			m.addActionListener((ActionEvent evt) -> {
				layoutLoadFile(mainFrame, key);
			});
			menu.add(m);
		}
		return (menu);
	}

	public static void layoutLoadFile(MainFrame mainFrame, String name) {
		//App.trace("DockUtil.LoadLayoutFile(mainframe," + name + ")");
		String x = "layout/" + name + LAYOUT_EXT;
		//App.trace("resource is " + x+"\n");
		try {
			InputStream stream = MainResources.class.getResourceAsStream(x);
			BufferedInputStream buf = new BufferedInputStream(stream);
			byte[] b = new byte[stream.available()];
			int r = buf.read(b);
			if (b != null && r > 0) {
				try ( ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(b))) {
					RootWindow rootWindow = mainFrame.getRootWindow();
					rootWindow.read(in, true);
				}
				mainFrame.cursorSetWaiting();
				setRespectMinimumSize(mainFrame);
				unloadHiddenViews(mainFrame);
				mainFrame.cursorSetDefault();
			}
		} catch (IOException ex) {
			LOG.err("unable to load docking resource : " + x);
		}
	}
}
